#include "stdafx.h"
#include "init.h"
#include <aux_klib.h>
#include "Undeclared.h"

NTSTATUS InitializeEnvironment(PUNICODE_STRING pusRegPath)
{
	ASSERT(pusRegPath && "Invalid base registry path passed");
	PAGED_CODE();
	// figure out the size of the buffer we need to store the full path
	// to out parameters key
	UNICODE_STRING usParamString = RTL_CONSTANT_STRING(L"\\Parameters");
	USHORT wBytesRequired = (pusRegPath->Length + usParamString.Length) + sizeof(WCHAR);
	ScopedArray<WCHAR> wszParamKeyPathBuffer(new WCHAR[wBytesRequired / sizeof(WCHAR)]);
	if(!*wszParamKeyPathBuffer)
	{
		DBG_TEXT("No memory for registry key buffer");
		return STATUS_INSUFFICIENT_RESOURCES;
	}
	else
	{
		UNICODE_STRING usParamKeyPath = {0, wBytesRequired, *wszParamKeyPathBuffer};
		NTSTATUS stat;
		RegModuleInformation rmiWin32k = {0};

		// clear the buffer and create the full path
		memset(*wszParamKeyPathBuffer, 0, wBytesRequired);
		RtlAppendUnicodeStringToString(&usParamKeyPath, pusRegPath);
		RtlAppendUnicodeStringToString(&usParamKeyPath, &usParamString);

		// the %uS is safe here as we're running at PASSIVE_LEVEL
		DBG_TEXT("Full path to registry key is %wZ", &usParamKeyPath);

		stat = QueryRegistryParams(&usParamKeyPath, &rmiWin32k);
		if(!NT_SUCCESS(stat))
		{
			DBG_TEXT("QueryRegistryParameters failed with status 0x%x", stat);
			return stat;
		}

		// the original driver checks that win32k is loaded at the same
		// address as it was at installation time. With drivers now
		// participating in dynamic address space layout this
		// check is only likely to succeed before the user
		// restarts the computer, so we ellide that check
		//
		// that said, I'm not really sure why the properties of win32k
		// are checked here rather than in dheapmon.exe
		// 
		// map win32k into view
		WinType<PVOID> pMapping(NULL, &UnmapModule);
		UNICODE_STRING usWin32kPath = RTL_CONSTANT_STRING(L"\\SystemRoot\\System32\\win32k.sys");
		stat = MapModule(&usWin32kPath, &pMapping, NULL);
		if(!NT_SUCCESS(stat))
		{
			return STATUS_UNSUCCESSFUL;
		}
		// force mode specified
		else if(rmiWin32k.dwTimeStamp == 0 && rmiWin32k.dwChecksum == 0)
		{
			return STATUS_SUCCESS;
		}
		// then check it's checksum and timestamp
		PIMAGE_NT_HEADERS pNtHead = RtlImageNtHeader(*pMapping);
		if((pNtHead->FileHeader.TimeDateStamp != rmiWin32k.dwTimeStamp) || 
			(pNtHead->OptionalHeader.CheckSum != rmiWin32k.dwChecksum))
		{
			DBG_TEXT(
				"win32k.sys mismatch\nExpected a timestamp of %x and a checksum of %x "
				"but found a timestamp of %x and a checksum of %x",
				rmiWin32k.dwTimeStamp,
				rmiWin32k.dwChecksum,
				pNtHead->FileHeader.TimeDateStamp,
				pNtHead->OptionalHeader.CheckSum
				);
			stat = STATUS_INVALID_DEVICE_STATE;
		}
		// if we loaded the offset to the EResource object
		// fix it up to a proper pointer
		else if(rmiWin32k.ullEResourceOffset)
		{
			stat = FixupPtrOffsets(&rmiWin32k);
			if(NT_SUCCESS(stat))
			{
				g_ppEResource = reinterpret_cast<PERESOURCE*>(rmiWin32k.ullEResourceOffset);
			}
		}
		return stat;
	}
}

NTSTATUS QueryRegistryParams(PUNICODE_STRING pusRegPath, RegModuleInformation* prmiWin32k)
{
	ASSERT(pusRegPath && prmiWin32k && "Invalid null arguments passed");
	PAGED_CODE();
	// EResource (offset to grpresUser)
	// Checksum of win32k
	// Timestamp of win32k
	static const ULONG dwNumberOfRegistryParameters = 3;
	RTL_QUERY_REGISTRY_TABLE rqrtQueryTable[dwNumberOfRegistryParameters + 1] = {{0}};
	LONGLONG llNullDefault = 0;

	// the joys of c&p
	// filling out the details of the values we're wanting from the registry
	//
	// number 1. the address of the win32k!grpresUser object
	prmiWin32k->ullEResourceOffset = static_cast<ULONGLONG>(-8);
	rqrtQueryTable[0].Flags = RTL_QUERY_REGISTRY_REQUIRED | RTL_QUERY_REGISTRY_DIRECT;
	rqrtQueryTable[0].Name = L"EResource";
	rqrtQueryTable[0].EntryContext = &(prmiWin32k->ullEResourceOffset);
	rqrtQueryTable[0].DefaultType = REG_QWORD;
	rqrtQueryTable[0].DefaultLength = sizeof(prmiWin32k->ullEResourceOffset);
	rqrtQueryTable[0].DefaultData = &llNullDefault;

	// number 2, the win32k.sys checksum at time of install
	rqrtQueryTable[1].Flags = RTL_QUERY_REGISTRY_REQUIRED | RTL_QUERY_REGISTRY_DIRECT;
	rqrtQueryTable[1].Name = L"CheckSum";
	rqrtQueryTable[1].EntryContext = &prmiWin32k->dwChecksum;
	rqrtQueryTable[1].DefaultType = REG_DWORD;
	rqrtQueryTable[1].DefaultLength = sizeof(prmiWin32k->dwChecksum);

	// number 3, the win32k.sys time stamp at time of install
	rqrtQueryTable[2].Flags = RTL_QUERY_REGISTRY_REQUIRED | RTL_QUERY_REGISTRY_DIRECT;
	rqrtQueryTable[2].Name = L"TimeStamp";
	rqrtQueryTable[2].EntryContext = &prmiWin32k->dwTimeStamp;
	rqrtQueryTable[2].DefaultType = REG_DWORD;
	rqrtQueryTable[2].DefaultLength = sizeof(prmiWin32k->dwTimeStamp);

	// rqrtQueryTable[3] is left NULL to indicate end of buffer

	return RtlQueryRegistryValues(RTL_REGISTRY_ABSOLUTE, pusRegPath->Buffer, rqrtQueryTable, NULL, NULL);
}

NTSTATUS FixupPtrOffsets(RegModuleInformation* prmiWin32k)
{
	ASSERT(prmiWin32k && "Invalid pointer passed");
	// the original version used ZwQuerySystemInformation(SystemModuleInformation, ...)
	// to get the module list, we'll use AuxKlib to minimize the 
	// number of things that could be changed without our knowledge
	// and that we'd have to compensate for in future versions
	//
	// If you've aleady seen the use of ZwQuerySystemInformation in Process.cpp
	// then pretend you haven't :)

	NTSTATUS stat = AuxKlibInitialize();
	if(!NT_SUCCESS(stat))
	{
		DBG_TEXT("AuxKlibInitialize failed 0x%x", stat);
		return stat;
	}
	ULONG dwModuleBufferRequired = 0;
	AUX_MODULE_EXTENDED_INFO* pModuleInfo = NULL;
	// we need the extended version for the module names
	AuxKlibQueryModuleInformation(&dwModuleBufferRequired, sizeof(AUX_MODULE_EXTENDED_INFO), pModuleInfo);
	ScopedArray<UCHAR> saModData(new UCHAR[dwModuleBufferRequired]);
	if(*saModData)
	{
		// get the module list
		pModuleInfo = reinterpret_cast<AUX_MODULE_EXTENDED_INFO*>(*saModData);
		stat = AuxKlibQueryModuleInformation(&dwModuleBufferRequired, sizeof(AUX_MODULE_EXTENDED_INFO), pModuleInfo);
		if(!NT_SUCCESS(stat))
		{
			DBG_TEXT("AuxKlibQueryModuleInformation failed 0x%x", stat);
			return stat;
		}
		// if we got them, search for win32k.sys
		STRING sWin32k = RTL_CONSTANT_STRING("win32k.sys");
		ULONG dwNumModules = dwModuleBufferRequired / sizeof(*pModuleInfo);
		ULONG i = 0;
		for(; i < dwNumModules; ++i, ++pModuleInfo)
		{
			// we're only interested in the file name
			const UCHAR* pszModName = pModuleInfo->FullPathName + pModuleInfo->FileNameOffset;
			STRING sModuleName;
			RtlInitString(&sModuleName, reinterpret_cast<PCSZ>(pszModName));
			if(RtlEqualString(&sWin32k, &sModuleName, TRUE))
			{
				break;
			}
		}
		// if we found it, fixup the offsets
		// otherwise fail
		if(i != dwNumModules)
		{
			UCHAR* pWin32kBase = static_cast<UCHAR*>(pModuleInfo->BasicInfo.ImageBase);
			UCHAR* pWin32kEnd = pWin32kBase + pModuleInfo->ImageSize;
			UCHAR* pEResource = pWin32kBase + static_cast<ULONG_PTR>(prmiWin32k->ullEResourceOffset);
			// this is to try and filter out someone using dheapinst.exe directly
			// instead of using our '.com' version of the tool which convert the 
			// direct address to an offset
			// I know 0x80000000 might not be the highest user address but
			// there's no way win32k is going to be 2GB big
			ULONGLONG ullUsual32BitDivide = 0x80000000ULL;
			if(((pWin32kBase < pEResource) && (pEResource < pWin32kEnd)) &&
				(prmiWin32k->ullEResourceOffset < ullUsual32BitDivide)
			)
			{
				prmiWin32k->ullEResourceOffset = reinterpret_cast<ULONGLONG>(pEResource);
				DBG_TEXT("grpresUser is at %p", pEResource);
			}
			else
			{
				DBG_TEXT("pEResource (%p) is outside the range of win32k (%p - %p)", pEResource, pWin32kBase, pWin32kEnd);
				stat = STATUS_DRIVER_ENTRYPOINT_NOT_FOUND;
			}
		}
		else
		{
			DBG_TEXT("win32k.sys not loaded!");
			stat = STATUS_DRIVER_ENTRYPOINT_NOT_FOUND;
		}
	}
	else
	{
		DBG_TEXT("Couldn't allocate %d bytes for AuxKlibQueryModuleInformation", dwModuleBufferRequired);
		stat = STATUS_INSUFFICIENT_RESOURCES;
	}
	return stat;
}

NTSTATUS MapModule(PUNICODE_STRING pusModulePath, PVOID* pMappedAddress, PSIZE_T pViewSize)
{
	ASSERT(pusModulePath && pMappedAddress && "Invalid NULL argument passed");
	OBJECT_ATTRIBUTES objAttrs = RTL_CONSTANT_OBJECT_ATTRIBUTES(pusModulePath, OBJ_CASE_INSENSITIVE);
	IO_STATUS_BLOCK iosb = {0};
	HANDLE hModuleFile = NULL;
	NTSTATUS stat = ZwOpenFile(&hModuleFile, FILE_GENERIC_READ, &objAttrs, &iosb, FILE_SHARE_READ, FILE_NON_DIRECTORY_FILE | FILE_SYNCHRONOUS_IO_NONALERT);
	*pMappedAddress = NULL;
	if(!NT_SUCCESS(stat))
	{
		DBG_TEXT("Failed to map %wZ because of error 0x%x", pusModulePath, stat);
		return stat;
	}
	WinType<> moduleFileWrap(hModuleFile, &ZwClose);
	HANDLE hModuleSection = NULL;
	stat = ZwCreateSection(&hModuleSection, SECTION_ALL_ACCESS, NULL, NULL, PAGE_READONLY, SEC_COMMIT, hModuleFile);
	if(!NT_SUCCESS(stat))
	{
		DBG_TEXT("Failed to create section of %wZ because of error 0x%x", pusModulePath, stat);
		return stat;
	}
	SIZE_T stViewSize = 0;
	stat = ZwMapViewOfSection(hModuleSection, NtCurrentProcess(), pMappedAddress, 0, 0, NULL, &stViewSize, ViewUnmap, 0, PAGE_READONLY);
	if(!NT_SUCCESS(stat))
	{
		DBG_TEXT("Failed to map view of %wZ because of error 0x%x", pusModulePath, stat);
	}
	else if(pViewSize)
	{
		*pViewSize = stViewSize;
	}
	ZwClose(hModuleSection);
	return stat;
}

NTSTATUS UnmapModule(PVOID pAddress)
{
	return ZwUnmapViewOfSection(NtCurrentProcess(), pAddress);
}