#include "stdafx.h"
#include "ioctls.h"
#include "list.h"
#include "WinstaDesk.h"
#include "structs.h"
#include "DesktopHeapInfo.h"

// disable the warning
// C4509: nonstandard extension used: 'GetDesktopCount' uses SEH and 'exLock' has destructor
// the generated code does the right thing
#pragma warning(push)
#pragma warning(disable : 4509)
NTSTATUS GetDesktopCount(PIRP pIrp, ULONG_PTR* pulpBytesWritten)
{
	ASSERT(pIrp && pulpBytesWritten && "Invalid NULL Argument(s)");
	PAGED_CODE();
	PIO_STACK_LOCATION pIoStack = IoGetCurrentIrpStackLocation(pIrp);
	const ULONG dwInLen = pIoStack->Parameters.DeviceIoControl.InputBufferLength;
	DeskCountInput* pDeskCountIn = reinterpret_cast<DeskCountInput*>(pIrp->AssociatedIrp.SystemBuffer);
	NTSTATUS stat = STATUS_SUCCESS;

	// first validate the parameters
	if((dwInLen < sizeof(DeskCountInput))
	|| (pDeskCountIn->signature != dwDeskCountSignature)
	|| (pDeskCountIn->hWinSta == reinterpret_cast<HANDLE>(-1)))
	{
		DBG_TEXT(__FUNCTION__ " input parameter verification failure");
		return STATUS_INVALID_BUFFER_SIZE;
	}

	// next check that the EResource address we were given (if any) is valid
	PERESOURCE pResource = NULL;
	if(g_ppEResource)
	{
		if(!CheckEResourceValidity(&pResource))
		{
			DBG_TEXT("win32k!gpresUser (address %p) verification failure", g_ppEResource);
			return STATUS_DRIVER_INTERNAL_ERROR;
		}
	}

	// if we got the resource, acquire it for the duration of the call
	// otherwise just do the call
	if(pResource)
	{
		ExclusiveResourceLock exLock(pResource);
		__try
		{
			stat = GetDesktopHeapInfo(pDeskCountIn->hWinSta, pIoStack->DeviceObject);
		}
		__except(stat = GetExceptionCode(), EXCEPTION_EXECUTE_HANDLER)
		{
			// swallow any exception to ensure the resource is released
			DBG_TEXT("Caught exception 0x%x from GetDesktopHeapInfo", stat);
		}
	}
	else
	{
		stat = GetDesktopHeapInfo(pDeskCountIn->hWinSta, pIoStack->DeviceObject);
	}
	if(NT_SUCCESS(stat))
	{
		ULONG* pdwDeskCount = static_cast<ULONG*>(pIrp->AssociatedIrp.SystemBuffer);
		*pdwDeskCount = CountDeskListEntries();
		*pulpBytesWritten = sizeof(ULONG);
	}
	else
	{
		EmptyDeskList();
		*pulpBytesWritten = 0;
	}
	return stat;
}
#pragma warning(pop)

NTSTATUS GetDesktopInfo(PIRP pIrp, ULONG_PTR* pulpBytesWritten)
{
	PAGED_CODE();
	PIO_STACK_LOCATION pIoStack = IoGetCurrentIrpStackLocation(pIrp);
	const ULONG dwNumberOfDesktops = CountDeskListEntries();
	const ULONG dwOutLen = pIoStack->Parameters.DeviceIoControl.OutputBufferLength;
	const ULONG dwOutputSizeRequired = dwNumberOfDesktops * sizeof(DesktopHeapInfo);

	DesktopHeapInfo* pDeskHeapOut = static_cast<DesktopHeapInfo*>(pIrp->AssociatedIrp.SystemBuffer);

	// validate params
	if(dwOutLen < dwOutputSizeRequired)
	{
		DBG_TEXT(__FUNCTION__ " output buffer too small, needed %d but got %d", dwOutputSizeRequired, dwOutLen);
		// not my choice of return code
		return STATUS_DEVICE_BUSY;
	}
	CopyDesktopListToBuffer(pDeskHeapOut);
	*pulpBytesWritten = dwOutputSizeRequired;
	// the original driver put the allocated desktop structures into a free list
	// for later reuse, we just free them
	EmptyDeskList();
	return STATUS_SUCCESS;
}

// disable the warning
// C4509: nonstandard extension used: 'CheckPtrPtrValidity' uses SEH and 'mdlWrap' has destructor
// the generated code does the right thing
#pragma warning(push)
#pragma warning(disable : 4509)
BOOLEAN CheckPtrPtrValidity(PVOID* ppAddress)
{
	PAGED_CODE();
	ASSERT(ppAddress && "Invalid pointer passed");
	BOOLEAN bIsValid = FALSE;
	PMDL pMdl = IoAllocateMdl(ppAddress, sizeof(PVOID), FALSE, FALSE, NULL);
	if(pMdl)
	{
		WinType<PMDL, VOID> mdlWrap(pMdl, &IoFreeMdl);
		PEXCEPTION_POINTERS pExPtr = NULL;
		__try
		{
			MmProbeAndLockPages(pMdl, KernelMode, IoReadAccess);
			// if we're here a first dereference was successful
			DBG_TEXT("First dereference, %p can be referenced", ppAddress);
			MmUnlockPages(pMdl);
			MmPrepareMdlForReuse(pMdl);
			MmInitializeMdl(pMdl, *ppAddress, sizeof(PVOID));
			MmProbeAndLockPages(pMdl, KernelMode, IoReadAccess);
			// and a second dereference
			MmUnlockPages(pMdl);
			DBG_TEXT("Second dereference, %p can be referenced", *ppAddress);
			bIsValid = TRUE;
		}
		__except(pExPtr = GetExceptionInformation(), EXCEPTION_EXECUTE_HANDLER)
		{
			DBG_TEXT(
				"Caught exception 0x%x at %p in "__FUNCTION__,
				pExPtr->ExceptionRecord->ExceptionCode,
				pExPtr->ExceptionRecord->ExceptionAddress
				);
		}
	}
	else
	{
		DBG_TEXT("No memory for MDL in "__FUNCTION__);
	}
	return bIsValid;
}
#pragma warning(pop)

BOOLEAN CheckEResourceValidity(PERESOURCE* ppEResource)
{
	PAGED_CODE();
	ASSERT(ppEResource && "Invalid NULL pointer passed in");
	BOOLEAN bIsEResourceValid = FALSE;
	if(CheckPtrPtrValidity(reinterpret_cast<PVOID*>(g_ppEResource)))
	{
		*ppEResource = *g_ppEResource;

		// this is all seems to be just checking whether the EResource is correctly
		// inserted into the list of system resources
		PERESOURCE pEResource = *ppEResource;
		PLIST_ENTRY pNextResource = pEResource->SystemResourcesList.Flink;
		PLIST_ENTRY pSecondResource = pNextResource->Flink;
		if(pSecondResource->Blink == pNextResource)
		{
			PLIST_ENTRY pBeforeNextResource = pNextResource->Blink;
			bIsEResourceValid = (pBeforeNextResource->Flink == pNextResource);
		}
	}
	return bIsEResourceValid;
}
